#!/usr/bin/env python3
# -*- encoding:utf-8 -*-

# Copyright (C) 2009-2016  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import argparse
import errno
import itertools
import logging
import Mimeo
import os
import shutil
import sqlite3
import sys
import time
import xdg.BaseDirectory
import xdg.IconTheme
import xml.dom.minidom



################################### Globals ####################################

NAME = 'ObFileBrowser'
LCNAME = NAME.lower()
DBNAME = '{}.sqlite'.format(LCNAME)

PYTHON_EXE = 'python3'

UNKNOWN_MIMETYPE = 'unknown/mimetype'



################################## Functions ###################################

def relevant_mtimes(mimeo, *paths):
  '''
  Return all paths that should be used to determine if the cached menus are
  still valid.
  '''
  for p in itertools.chain.from_iterable((
    mimeo.mimeapps_directories(),
    mimeo.mimeapps_list_paths(),
    mimeo.desktop_paths(),
    mimeo.mimeinfo_caches(),
    paths
  )):
    try:
      yield os.path.getmtime(p)
    except FileNotFoundError:
      pass



def add_icon(element, icon):
  '''
  Add an icon attribute to the given element.
  '''
  if icon:
    element.setAttribute('icon', xdg.IconTheme.getIconPath(icon))



################################### PipeMenu ###################################

class PipeMenu(object):

  def __init__(self, impl, element="openbox_pipe_menu"):
    '''
    impl: xml.dom.minidom.getDOMImplementation
    '''
    self.doc = impl.createDocument(None, element, None)



  def __str__(self):
    return self.doc.documentElement.toprettyxml(indent='  ')



  # Escape underscores.
  def escape_label(self, label):
    return label.replace('_', '__')



  def clear(self):
    for child in self.doc.documentElement.childNodes:
      self.doc.documentElement.removeChild(child)



  def add_separator(self, label=None):
    sep = self.doc.createElement('separator')
    if label:
      sep.setAttribute('label', self.escape_label(label))
    self.doc.documentElement.appendChild(sep)



  # An empty menu will be replaced by one with the matching ID in Openbox.
  def add_menu(self, ident, label, menu=None, icon=None):
    if menu:
      if isinstance(menu, xml.dom.minidom.Document):
        menu = menu.documentElement

      elif isinstance(menu, PipeMenu):
        menu = menu.doc.documentElement

      menu.tagName = 'menu'
      if menu.hasAttributes():
        for attr in menu.attributes:
          menu.removeAttribute(attr)
    else:
      menu = self.doc.createElement('menu')
    menu.setAttribute('id', 'menu: {}'.format(ident))
    menu.setAttribute('label', self.escape_label(label))
    add_icon(menu, icon)
    self.doc.documentElement.appendChild(menu)



  def add_pipe_menu(self, ident, label, execute, icon=None):
    menu = self.doc.createElement('menu')
    menu.setAttribute('id', 'pipemenu: {}'.format(ident))
    menu.setAttribute('label', self.escape_label(label))
    menu.setAttribute('execute', execute)
    add_icon(menu, icon)
    self.doc.documentElement.appendChild(menu)



  def add_command(self, label, command_str, startup_notify=False, icon=None):
    item = self.doc.createElement('item')
    item.setAttribute('label', self.escape_label(label))
    add_icon(item, icon)
    action = self.doc.createElement('action')
    action.setAttribute('name', 'Execute')
    command = self.doc.createElement('command')
    cmd = self.doc.createTextNode(command_str)

    command.appendChild(cmd)
    action.appendChild(command)

    # startup notification
    if startup_notify:
      startupnotify = self.doc.createElement('startupnotify')
      enabled = self.doc.createElement('enabled')
      yes = self.doc.createTextNode('yes')
      enabled.appendChild(yes)
      startupnotify.appendChild(enabled)
      action.appendChild(startupnotify)

    item.appendChild(action)
    self.doc.documentElement.appendChild(item)



  def add_text(self, text, icon=None):
    item = self.doc.createElement('item')
    item.setAttribute('label', self.escape_label(text))
    add_icon(item, icon)
    self.doc.documentElement.appendChild(item)



################################ ObFileBrowser #################################

class ObFileBrowser():
  def __init__(
    self,
    impl,
    mimeo,
    cmd,
    show_hidden=True,
    show_parent=True,
    show_self=True,
    commands=list()
  ):
    '''
    impl: xml.dom.minidom.getDOMImplementation
    mimeo: instance of Mimeo
    cmd:
    show_hidden: show hidden (dot) files
    show_parent: allow navigation to parent
    show_self:
    commands:
    '''
    self.impl = impl
    self.mimeo = mimeo
    self.cmd = cmd
    self.show_hidden = show_hidden
    self.show_parent = show_parent
    self.show_self = show_self
    self.commands = commands


  def parse_command(self, command, fpath):

    while '%FILEPATH%' in command:
      command = command.replace('%FILEPATH%', fpath)

    if '%MIMETYPE%' in command:
      try:
        mimetype = self.mimeo.arg_to_mimetypes(fpath)
      except StopIteration:
        mimetype = UNKNOWN_MIMETYPE
      while '%MIMETYPE%' in command:
        command = command.replace('%MIMETYPE%', mimetype)

    return command



  def launcher_menu(self, fpath):
    mimeo = self.mimeo
    is_empty = True
    add_sep = True

    menu = PipeMenu(self.impl)
    for a, de in mimeo.args_to_desktop_entries((fpath,)):
      # There should be only 1.
      for cmd in Mimeo.desktop_entry_to_cmds(de, (a,)):
        menu.add_command(de.getName(), Mimeo.quote_cmd(cmd), de.getStartupNotify(), icon=de.getIcon())
        is_empty = False

    if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
      menu.add_separator(label='Run')
      menu.add_command('Directly', fpath)
      if mimeo.term_cmd:
        menu.add_command('In terminal', Mimeo.quote_cmd(Mimeo.interpolate_term_cmd(mimeo.term_cmd, (fpath,))))
      is_empty = False

    if self.commands:
      menu.add_separator(label='Custom')
      for name, command in self.commands:
        command = self.parse_command(command, fpath)
        menu.add_command(name, command)
        is_empty = False

    if is_empty:
      menu.add_separator(label='Error')
      menu.add_text('Failed to determine launchers')
      menu.add_separator(label='Debug')
      menu.add_text('Check the output of:')
      mimeo_cmd = ('mimeo', '--desktop', fpath)
      mc_label = Mimeo.quote_cmd(mimeo_cmd)
      if mimeo.term_cmd:
        menu.add_command(
          mc_label,
          Mimeo.quote_cmd(Mimeo.interpolate_term_cmd(mimeo.term_cmd, mimeo_cmd)))
      else:
        menu.add_text(mc_label)
    return menu



  def pipe_menu(self, path):
    path = os.path.abspath(path)
    menu = PipeMenu(self.impl)
    menu.add_separator(label=path)

    # Track additions that require a separator if the following item is present.
    add_sep = False

    if os.path.isdir(path):

      # directories, hidden directories, files, hidden files
      dirs = list()
      hdirs = list()
      files = list()
      hfiles = list()

      try:
        for item in sorted(os.listdir(path)):
          fpath = os.path.join(path, item)

          if item.startswith('.'):
            if not self.show_hidden:
              continue
            elif os.path.isdir(fpath):
              hdirs.append((fpath, item))
            else:
              hfiles.append((fpath, item))
          elif os.path.isdir(fpath):
            dirs.append((fpath, item))
          else:
            files.append((fpath, item))
      # Ignore paths that we cannot read.
      except OSError as e:
        if e.errno == errno.EACCES:
          logging.warning('failed to read {} [{}]'.format(path, e))
        else:
          raise e

      if self.show_parent or True:
        parent = os.path.abspath(os.path.join(path, '..'))
        if parent != path:
          cmd = Mimeo.quote_cmd(self.cmd + [parent])

          pname = os.path.basename(parent)
          if not pname:
            pname = '/'
          menu.add_pipe_menu(parent, pname, cmd)
          add_sep = True

      if self.show_self:
        menu.add_menu(path, '.', self.launcher_menu(path))
        add_sep = True

      if self.show_hidden and (hdirs or hfiles):
        hidden = PipeMenu(self.impl)
        hidden.add_separator(label="Hidden")

        if hdirs:
          for fpath, item in hdirs:
            cmd = Mimeo.quote_cmd(self.cmd + [fpath])
            hidden.add_pipe_menu(fpath, item, cmd)

        if files:
          if hdirs:
            menu.add_separator()
          for fpath, item in hfiles:
            hidden.add_menu(fpath, item, self.launcher_menu(fpath))

        menu.add_menu('hidden: {}'.format(path), 'hidden', hidden)

      if dirs:
        menu.add_separator(label='Dirs')
        add_sep = True
        for fpath, item in dirs:
          cmd = Mimeo.quote_cmd(self.cmd + [fpath])
          menu.add_pipe_menu(fpath, item, cmd)

      if files:
        menu.add_separator(label='Files')
        add_sep = True
        for fpath, item in files:
          menu.add_menu(fpath, item, self.launcher_menu(fpath))

    else:
      menu.add_menu(path, os.path.basename(path), self.launcher_menu(path))

    return menu





############################ ObDesktopEntryBrowser #############################

class ObDesktopEntryBrowser(object):
  def __init__(self, impl, mimeo, obfb_cmd, load=False):
    self.impl = impl
    self.mimeo = mimeo
    self.id = 'ObMimeoDesktopBrowser'
    self.seen = set()
    self.menu = PipeMenu(self.impl)
    self.obfb_cmd = obfb_cmd

    if load:
      self.load()



  def __str__(self):
    return str(self.menu)


  def get_key(self, tup):
    return tup[0].lower()



  def load(self):
    self.menu.clear()

    names = list()
    cats = dict()

    for desktop_path in self.mimeo.desktop_paths():
      de = Mimeo.desktop_entry(desktop_path, none_if_error=True)
      if not de or de.getNoDisplay():
        continue

      name = de.getName()
      pair = (name, de)
      names.append(pair)

      for cat in de.getCategories():
        try:
          cats[cat].append(pair)
        except KeyError:
          cats[cat] = [pair]

    self.names = sorted(names, key=self.get_key)
    for cat in cats:
      cats[cat] = sorted(cats[cat], key=self.get_key)
    self.cats = cats



    # Sorted by name.
    name_menu = PipeMenu(self.impl)
    name_menu.add_separator(label='Names')

    subname_menu = None
    c = None
    for name, de in self.names:
      d = name[0].lower()
      if d != c:
        # Add the previous submenu.
        if subname_menu:
          name_menu.add_menu('{}: letter {}'.format(self.id, c), c, subname_menu)

        # Create a new submenu for this letter.
        subname_menu = PipeMenu(self.impl)
        subname_menu.add_separator(label=d)
        c = d

      dmenu = self.desktop_launcher(de)
      subname_menu.add_menu('{}: {}'.format(self.id, de.filename), name, menu=dmenu, icon=de.getIcon())

    if subname_menu:
      name_menu.add_menu('{}: letter {}'.format(self.id, c), c, subname_menu)

    # Sorted by category
    cat_menu = PipeMenu(self.impl)
    cat_menu.add_separator(label='Categories')
    for cat, des in sorted(self.cats.items(), key=self.get_key):
      menu = PipeMenu(self.impl)
      menu.add_separator(label=cat)
      for name, de in des:
        dmenu = self.desktop_launcher(de)
        menu.add_menu('{}: {}'.format(self.id, de.filename), name, menu=dmenu, icon=de.getIcon())
      cat_menu.add_menu('{}: category {}'.format(self.id, cat), cat, menu)

    self.menu.add_separator(label='Desktop Entries')
    self.menu.add_menu('{}: Names'.format(self.id), 'Names', name_menu)
    self.menu.add_menu('{}: Categories'.format(self.id), 'Categories', cat_menu)



  def desktop_launcher(self, de):
    fpath = de.filename
    menu = PipeMenu(self.impl)

    if not fpath in self.seen:
      name = de.getName()
      gen_name = de.getGenericName()
      comment = de.getComment()
      startup_notify = de.getStartupNotify()
      term = de.getTerminal()
      icon = de.getIcon()
      cats = de.getCategories()

      menu.add_separator(label=name)

      try:
        cmd = next(Mimeo.desktop_entry_to_cmds(de, term_cmd=self.mimeo.term_cmd))
      except StopIteration:
        return PipeMenu
      cmd = Mimeo.quote_cmd(cmd)
      if term and not self.mimeo.term_cmd:
        menu.add_text('error: terminal required')
      else:
        menu.add_command(cmd, cmd, startup_notify=startup_notify, icon=icon)

      menu.add_separator(label='Description')
      if gen_name:
        menu.add_text(gen_name)
      if comment:
        menu.add_text(comment)

      if cats:
        menu.add_separator(label='Categories')
        for cat in sorted(cats):
          menu.add_text(cat)

      menu.add_separator(label='Location')
      obfb_cmd = Mimeo.quote_cmd(self.obfb_cmd + [fpath])
      menu.add_pipe_menu('{}: launcher {}'.format(self.id, fpath), fpath, obfb_cmd)

      self.seen.add(fpath)

    return menu



############################### Argument Parsing ###############################

class DisplayCommandHelp(argparse.Action):
  def __call__(self, parser, namespace, values, option_string=None):
    print('''OPENBOX MENU EXAMPLE
  The following line can be added to an Openbox menu in menu.xml to include an
  ObFileBrowser menu that opens the user's home directory and uses urxvtc to
  open any terminal applications.

  <menu id="obfb-home" label="home" execute="obfilebrowser ~ --term 'urxvtc -e %s'" />



COMMANDS
  Custom commands can be added to the Exec menu for files and directories using
  "--command <name> <command>". This may be expanded in the future, depending
  on feedback.

  <name>
    The name to display in the menu.

  <command>
    The command to execute. Substrings will be replaced as follows:

      %FILEPATH%
        The full filepath.

      %MIMETYPE%
        The detected mimetype.

  Examples
    Move files to a trash directory:
      --command trash 'mv "%FILEPATH%" -t /path/to/trash'

    Copy the file path to the clipboard:
      --command 'copy path' 'sh -c "echo \"%FILEPATH%\" | xsel -ib"'

    Note the use of "sh -c" due to the pipe. Openbox does not recognize pipes
    and other redirections so they must be wrapped.
''')
    sys.exit(0)



def get_parser(parser=None, prog=NAME.lower()):

  if not parser:
    parser = argparse.ArgumentParser(
      prog=prog,
      description='Browse and open files using Openbox pipe menus.'
    )

  parser.add_argument(
    'fpath', metavar='<filepath>', default='.', nargs='?',
    help='The path to a file or directory.'
  )

  parser.add_argument(
    '--no-parent', dest='parent', action='store_false',
    help='Remove the ".." entry from the initial directory.'
  )

  parser.add_argument(
    '--no-self', dest='self', action='store_false',
    help='Remove the "." launcher menu from the initial directory.'
  )

  parser.add_argument(
    '--no-hidden', dest='hidden', action='store_false',
    help='Remove the "hidden" menu containing hidden files and directories.'
  )

  parser.add_argument(
    '--term', metavar='<cmd>', action='store',
    help='Specify a terminal command. See Mimeo\'s documentation for details.'
  )

  parser.add_argument(
    '--command', dest='commands', default=list(), nargs=2, metavar=('<name>', '<command>'), action='append',
    help='Custom commands to add to every Exec menu. See "--more-help" for details.'
  )

  parser.add_argument(
    '--more-help', nargs=0, action=DisplayCommandHelp,
    help='Display more information.'
  )

  parser.add_argument(
    '--desktop-menu', action='store_true',
    help='Display desktop entries sorted by name and category.'
  )

  parser.add_argument(
    '--cache', action='store_true',
    help='Cache menus in a database to speed up subsequent browsing. This should detect changes in target directories and MIME-type associations.'
  )

  parser.add_argument(
    '--delete-cache', action='store_true',
    help='Delete the cache database.'
  )

  return parser



##################################### Main #####################################

def main(args=None):
  pargs = get_parser().parse_args(args)

  # For recursive invocations.
  sys.argv[0] = os.path.abspath(sys.argv[0])
  if os.access(sys.argv[0], os.X_OK):
    cmd = [sys.argv[0]]
    paths = [sys.argv[0]]
  else:
    cmd = [PYTHON_EXE, sys.argv[0]]
    paths = [Mimeo.which(PYTHON_EXE), sys.argv[0]]

  # Add arguments to recursive invocation command.
  if not pargs.parent:
    cmd.append('--no-parent')

  if not pargs.self:
    cmd.append('--no-self')

  if not pargs.hidden:
    cmd.append('--no-hidden')

  if pargs.term:
    cmd.append('--term')
    cmd.append(pargs.term)

  for name, command in pargs.commands:
    cmd.append('--command')
    cmd.append(name)
    cmd.append(command)

  if pargs.cache:
    cmd.append('--cache')

  mimeo = Mimeo.Mimeo(
    include_deprecated=True,
    term_cmd=pargs.term
  )

  cache = xdg.BaseDirectory.save_cache_path(LCNAME)
  if pargs.delete_cache:
    try:
      shutil.rmtree(cache)
    except FileNotFoundError:
      pass
    except OSError as e:
      try:
        os.remove(cache)
      except OSError:
        raise e
    return

  dbpath = os.path.join(cache, DBNAME)

  menu_id = cmd[1:] if cmd[0] == PYTHON_EXE else cmd[:]
  menu_id.append(pargs.fpath)
  menu_id = Mimeo.quote_cmd(menu_id)

  conn = None
  c = None



  if pargs.cache:
    if pargs.fpath and not pargs.desktop_menu:
      paths.append(os.path.abspath(pargs.fpath))
    newest = max(relevant_mtimes(mimeo, *paths))
    newest = round(newest)
  else:
    newest = 0



  if pargs.cache:
    conn = sqlite3.connect(
      dbpath,
      detect_types=(sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES),
      isolation_level=None
    )
    conn.text_factory=str
    c = conn.cursor()
    c.execute('CREATE TABLE IF NOT EXISTS menus (id TEXT PRIMARY KEY, xml TEXT, mtime INTEGER)')
    c.execute('DELETE FROM menus WHERE mtime<=?', (newest,))
    conn.commit()
    c.execute('select xml from menus where id=?', (menu_id,))
    row = c.fetchone()
    if row:
      print(row[0])
      return



  impl = xml.dom.minidom.getDOMImplementation()
  if pargs.desktop_menu:
    menu_xml = ObDesktopEntryBrowser(impl, mimeo, cmd, load=True)
  else:
    menu_xml = ObFileBrowser(
      impl,
      mimeo,
      cmd,
      show_hidden=pargs.hidden,
      show_parent=pargs.parent,
      show_self=pargs.self,
      commands=pargs.commands
    ).pipe_menu(pargs.fpath)
  menu_xml = str(menu_xml)
  print(menu_xml)



  if pargs.cache:
    c.execute('INSERT INTO menus (id, xml, mtime) VALUES (?,?,?)', (menu_id, menu_xml, int(time.time())))
    c.close()
    conn.commit()



if __name__ == "__main__":
  try:
    main()
  except (KeyboardInterrupt, BrokenPipeError):
    pass
  except sqlite3.OperationalError as e:
    sys.exit('sqlite3.OperationalError: {}'.format(e))
